import pickle
import unittest

import numpy as np
import tensorflow as tf

from akro import Discrete
from akro import theano
from akro.requires import requires_tf, requires_theano


class TestDiscrete(unittest.TestCase):
    def test_pickleable(self):
        obj = Discrete(10)
        round_trip = pickle.loads(pickle.dumps(obj))
        assert round_trip
        assert round_trip.n == 10

    def test_flat_dim(self):
        disc = Discrete(10)
        assert disc.flat_dim == 10

    def test_flatten(self):
        disc = Discrete(10)
        x = [3, 5, 7]
        arr = disc.flatten(x)
        assert all(arr[x] == 1)

    def test_unflatten(self):
        disc = Discrete(10)
        x = [3, 5, 7]
        arr = disc.flatten(x)
        assert disc.unflatten(arr) == 3

    def test_flatten_n(self):
        disc = Discrete(3)
        obs = np.asarray([0, 1, 2])
        arr = disc.flatten_n(obs)
        base = np.asarray([[1., 0., 0.], [0., 1., 0.], [0., 0., 1.]])
        assert np.array_equal(arr, base)

    def test_unflatten_n(self):
        disc = Discrete(3)
        obs = np.asarray([0, 1, 2])
        flat_arr = disc.flatten_n(obs)
        base = np.asarray([0, 1, 2])
        unflat_arr = disc.unflatten_n(flat_arr)
        assert np.array_equal(unflat_arr, base)

    def test_weighted_sample(self):
        disc = Discrete(4)
        weights = [0.1, 0.2, 0.3, 0.4]
        res = disc.weighted_sample(weights)
        assert res >= 0 and res < disc.n

    def test_weighted_sample_unnormalized(self):
        disc = Discrete(4)
        weights = np.array([1., 2., 3., 5.])
        res = disc.weighted_sample(weights)
        assert res >= 0 and res < disc.n

    def test_hash(self):
        disc = Discrete(10)
        assert disc.__hash__() == 10

    @requires_tf
    def test_convert_tf(self):
        disc = Discrete(10)
        tensor = disc.to_tf_placeholder('test', 1)
        assert isinstance(tensor, tf.Tensor)
        assert disc.dtype == np.int64
        assert tensor.dtype == tf.int64
        assert tensor.get_shape().as_list() == [None, 10]

    @requires_theano
    def test_convert_theano(self):
        disc = Discrete(10)
        tensor = disc.to_theano_tensor('test', 1)
        assert isinstance(tensor, theano.tensor.TensorVariable)
        assert disc.dtype == np.int64
        assert tensor.dtype == 'int64'



"""The abstract base class for all Space types."""

import abc

import gym.spaces


class Space(abc.ABC, gym.spaces.Space):
    """Provides a classification state spaces and action spaces.

    Allows you to write generic code that applies to any Environment.
    E.g. to choose a random action.
    """

    @abc.abstractmethod
    def flatten(self, x):
        """Return a flattened observation x.

        Args:
            x (:obj:`Iterable`): The object to flatten.

        Returns:
            np.ndarray: An array of x collapsed into one dimension.

        """

    @abc.abstractmethod
    def unflatten(self, x):
        """Return an unflattened observation x.

        Args:
            x (:obj:`Iterable`): The object to unflatten.

        Returns:
            np.ndarray: An array of x in the shape of self.shape.

        """

    @abc.abstractmethod
    def flatten_n(self, xs):
        """Return flattened observations xs.

        Args:
            xs (:obj:`Iterable`): The object to reshape and flatten

        Returns:
            np.ndarray: An array of xs in a shape inferred by the size of
                its first element.

        """

    @abc.abstractmethod
    def unflatten_n(self, xs):
        """Return unflattened observations xs.

        Args:
            xs (:obj:`Iterable`): The object to reshape and unflatten

        Returns:
            np.ndarray: An array of xs in a shape inferred by the size of
                its first element and self.shape.

        """

    @property
    def flat_dim(self):
        """Return the length of the flattened vector of the space."""

    @abc.abstractmethod
    def to_tf_placeholder(self, name, batch_dims):
        """Create a tensor placeholder from the Space object.

        Args:
            name (str): name of the variable
            batch_dims (:obj:`list`): batch dimensions to add to the
                shape of the object.

        Returns:
            tf.Tensor: Tensor object with the same properties as
                the Dict where the shape is modified by batch_dims.

        """

    @abc.abstractmethod
    def to_theano_tensor(self, name, batch_dims):
        """Create a theano tensor from the Space object.

        Args:
            name (str): name of the variable
            batch_dims (:obj:`list`): batch dimensions to add to the
                shape of the object.

        Returns:
            theano.tensor.TensorVariable: Tensor object with the
                same properties as the Dict where the shape is
                modified by batch_dims.

        """



import pickle
import unittest

import numpy as np
import tensorflow as tf

from akro import Box
from akro import Discrete
from akro import theano
from akro import Tuple
from akro.requires import requires_tf, requires_theano


class TestTuple(unittest.TestCase):
    def test_pickleable(self):
        tup = Tuple((Discrete(3), Discrete(2)))
        round_trip = pickle.loads(pickle.dumps(tup))
        assert round_trip
        assert round_trip.spaces == tup.spaces

    def test_flat_dim(self):
        tup = Tuple((Discrete(3), Discrete(2)))
        assert tup.flat_dim == 5

    def test_flatten(self):
        tup = Tuple((Discrete(3), Discrete(2)))
        x = [2, 0]
        arr = tup.flatten(x)
        assert arr[2] == arr[3] == 1

    def test_unflatten(self):
        tup = Tuple((Discrete(3), Discrete(2)))
        x = [2, 0]
        arr = tup.flatten(x)
        assert tup.unflatten(arr) == (2, 0)

    def test_flatten_n(self):
        disc = Discrete(3)
        tup = Tuple((Discrete(2), disc))
        obs = disc.flatten_n(np.asarray([0, 1, 0, 1, 2]))
        unflat_ret = tup.unflatten_n(obs)
        flat_ret = tup.flatten_n(unflat_ret)
        base = np.asarray([[1., 0., 1., 0., 0.]])
        assert np.array_equal(flat_ret, base)

    def test_unflatten_n(self):
        disc = Discrete(3)
        tup = Tuple((Discrete(2), disc))
        obs = disc.flatten_n(np.asarray([0, 1, 0, 1, 2]))
        ret = tup.unflatten_n(obs)
        assert ret == [(0, 0)]

    def test_hash(self):
        tup = Tuple((Discrete(3), Discrete(2)))
        assert tup.__hash__() == 3713083796995235906

    @requires_tf
    def test_convert_tf(self):
        tup = Tuple((Box(0.0, 1.0, (3, 4)), Discrete(2)))
        tensor_tup = tup.to_tf_placeholder('test', 1)
        assert isinstance(tensor_tup, tuple)
        assert all([isinstance(c, tf.Tensor) for c in tensor_tup])
        assert [c.dtype for c in tensor_tup] == [tf.float32, tf.int64]
        assert [c.get_shape().as_list() for c in tensor_tup] == [[None, 3, 4],
                                                                 [None, 2]]

    @requires_theano
    def test_convert_theano(self):
        tup = Tuple((Box(0.0, 1.0, (3, 4)), Discrete(2)))
        tensor_tup = tup.to_theano_tensor('test', 1)
        assert isinstance(tensor_tup, tuple)
        assert all(
            [isinstance(c, theano.tensor.TensorVariable) for c in tensor_tup])
        assert [c.dtype for c in tensor_tup] == ['float32', 'int64']



"""akro setuptools script."""
from setuptools import find_packages
from setuptools import setup

# Required dependencies
required = [
    # Please keep alphabetized
    'gym>=0.12.4',
    'numpy',
]

# Framework-specific dependencies
extras = {
    'tf': ['tensorflow<2.0'],
    'theano': ['theano'],
}
extras['all'] = list(set(sum(extras.values(), [])))

# Development dependencies (*not* included in "all")
extras['dev'] = [
    # Please keep alphabetized
    'coverage',
    'flake8',
    'flake8-docstrings>=1.5.0',
    'flake8-import-order',
    'pep8-naming',
    'pre-commit',
    'pycodestyle>=2.5.0',
    'pydocstyle>=4.0.0',
    'pylint',
    'pytest>=4.4.0',  # Required for pytest-xdist
    'pytest-cov',
    'pytest-xdist',
    'sphinx',
    'recommonmark',
    'yapf',
]

with open('README.md') as f:
    readme = f.read()

# Get the package version dynamically
with open('VERSION') as v:
    version = v.read().strip()

setup(
    name='akro',
    version=version,
    author='Reinforcement Learning Working Group',
    author_email='akro@noreply.github.com',
    description='Spaces types for reinforcement learning',
    url='https://github.com/rlworkgroup/akro',
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    python_requires='>=3.5',
    install_requires=required,
    extras_require=extras,
    license='MIT',
    long_description=readme,
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Software Development :: Libraries',
    ],
)
